class BracketParser 

  class NoToken < Exception ; end
  class InvalidFormat < Exception ; end 

  attr_reader :tokens, :left, :right

  def initialize(lb, rb, treetype)
    @stack = []
    @treetype = treetype
    @state = 'empty'
    @trace = false
    @left = lb
    @right = rb
  end

  def str2tokens(str, delimitor=/^ +$/)
    tokens = []
    str.split(/(#{@left.source}|#{@right.source}|#{delimitor.source})/).each do |x| 
      tokens << x if x != "" &&  delimitor !~ x
    end
    return tokens 
  end

  def tokenize(str=nil, delimitor=/^ +$/)
    return str2tokens(str, delimitor)
  end

  def trace_on
    @trace = true
  end

  def trace_off
    @trace = false
  end

  def to_s
    return '[state:%8s]|%s' % [@state, @stack.join('|')]
  end

  def push
    @state = 'push'
    @stack << @treetype.new
    
  end

  def pop
    if @stack.size == 0 
      raise InvalidFormat, "too many %s#{@right.source}%s" % ['\'', '\'']
    end
    @state = 'reduce'
    return @stack.pop
  end


  def add
    @state = 'add'
  end

  def parse(tokens)
    if tokens.size == 0 
      raise NoToken, "input token is empty" 
    end

    token = tokens[0]

    if @left =~ token 
      self.push 
    elsif @right =~ token 
      #puts @stack[0].parent.type
      n = self.pop 
      
      if @stack.size == 0 
	return n
      end
      
      @stack[-1 + @stack.size].insert(n) if n != nil

    else
      self.add
      @stack[-1 + @stack.size].insert(@treetype.new(token))
    end

    tokens.shift
    puts to_s if @trace
    return parse(tokens)
  end
end

if __FILE__ == $0
  require 'tree'
  require 'tree_viewer'
  bp = BracketParser.new(/\(/, /\)/, Tree.new.type)

  #bp = BracketParser.new(/\[/, /\]/)
  #bp = BracketParser.new(/\[/, /\]/, MyTree.new.type)

  lines = []

  while line = $stdin.gets
    line.chomp!
    
    next if /^[\s\t]*$/ =~ line

    if /^\(/ =~ line && lines.size != 0
      break 
    end
    lines << line 
  end

  tokens = bp.tokenize(lines.join(' '))
  # puts bp.tokens
  bp.trace_on

  tree = bp.parse(tokens)

  puts lines.join('')
  puts tree.bracket('(', ')', '')
  exit

  tv = TreeViewer.new(tree)
  tv.show
  #puts tree.to_s
  #puts tree.bracket
  #tree.show

  #tree.traverse do |x|
    #puts "===============" 
    #puts x.show
    #puts "===============" 
  #end

  #tree = bp.parse()
  
  #puts tree.bracket
  #p bp.tokens
  #tree = bp.parse()
  # p bp.tokens
  #p tree 
  #puts tree.to_s
  #puts tree.bracket
  #tree.show_tree
  exit
end
