require 'tree'
require 'pos'

class DependencyTree < Tree

  attr_accessor :label, :sid, :pos, :lex, :semantic, :chunk, :relation

  def initialize(data=nil, parent=nil, prev=nil, _next=nil, children=[])
    super( data=nil, parent=nil, prev=nil, _next=nil, children=[])
    @pos  = []
    @lex  = nil
    @semantic = nil
    @chunk = nil 
    @modifies = Hash.new
    @label = nil
    @relation = nil
  end

  def depth(_root = nil)
    return 0 if @parent == nil || self == _root
    return @parent.depth(_root) + 1
  end

  def is_root?
    return @parent == nil 
  end

  def is_leaf?
    return num_of_children == 0 
  end

  def each_child
    @children.each do |_ch|
      yield(_ch)
    end
  end

  def DependencyTree.gap
    gn = DependencyTree.new
    gn.pos << PartOfSpeech.gap
    gn.lex = GAP.downcase
    return gn
  end

  def DependencyTree.bos
    bn = DependencyTree.new
    bn.pos << PartOfSpeech.bos
    bn.sid = -1
    return bn
  end

  def set_pos(str, delimitor =/\|/)
    str.split(delimitor).each do |_pos|
      @pos << PartOfSpeech.str2pos(_pos)
    end
  end

  def each_pos
    @pos.each do |_pos|
      yield(_pos)
    end
  end


  def DependencyTree.eos
    en = DependencyTree.new
    en.pos << PartOfSpeech.eos
    return en
  end

  def eos?
    return @pos[0].eos?
  end

  def bos?
    return @pos[0].bos?
  end


  def gap?
    return @pos[0].gap?
  end

  def best_pos
    return @pos[0]
  end

  def modifier
    return @parent
  end
  
  def modify
    return @children
  end

  def top?
    return root?
  end

  def bracket(left=@@left, right=@@right, delimitor=@@delimitor, sem=false, h=false)
    return "%s%s%s" % [left, self.to_s,  right]  if leaf?
    tmp = [] 
    tmp << self.to_s
    @children.each do |x|
      tmp << x.bracket(left, right, delimitor, sem, h)
    end
    return "%s%s%s" % [left, tmp.join(delimitor), right]
    
  end

  def to_s
    return "%d:%s/%s" % [@sid, @lex, @pos[0], @label, @semantic] 
  end

  def to_list(rid=-1)
    list = []
    terminals.each do |x|
      next if x.bos? || x.eos?
      tmp = []
      tmp << x.lex
      tmp << x.pos

      if x.root? || x.parent.sid == nil 
	tmp << rid
      else
	tmp << x.parent.sid 
      end
      if x.chunk != nil 
	tmp << x.chunk 
      else
	tmp << "O"
      end
      tmp << x.label if x.label != nil 
      tmp << x.semantic if x.label != nil 
      is_leaf = x.leaf? 

      tmp << is_leaf
      list << tmp.join("\t")
    end
    return list
  end

  def each 
    terminals.each do |x|
      yield(x) 
    end
  end

  def num_of_children
    return @children.size
  end

  def num_of_left_children
    num = 0 
    @children.each do |x|
      num += 1 if x.offset <  @offset
    end
    return num 
  end

  def num_of_right_children
    num = 0 
    @children.each do |x|
      num += 1 if x.offset >  @offset
    end
    return num 
  end

  def show(os=$stdout)
    maxdepth = 0 
    nodes = []
    d2ml = Hash.new(0)
    
    depth_first_traverse_pre_order do |n|
      d = n.depth(self)
      d2ml[d] = n.to_s.size  if d2ml[d] < n.to_s.size  
      maxdepth = d if d > maxdepth
      nodes << n
    end

    o2l = Hash.new
    l = 0 
    nodes.sort{|x, y| x.offset <=> y.offset}.each do |n| 
      o2l[n.offset] = l 
      l += 1
    end

    vm = Array.new(nodes.size * 2) 
    vm.each_index do |i|
      vm[i] = Array.new(maxdepth  + 1)
      vm[i].each_index do |j|
	vm[i][maxdepth - j] = " " *  (d2ml[j] + 2)
      end
    end

    depth_first_traverse_pre_order do |n|
      d = n.depth(self)
      m = '+'

      m = '=' if n.is_root?

      vm[ o2l[n.offset] * 2 ][maxdepth - d] = "%s%s%s" % [n, "-" * (d2ml[d] + 1 - n.to_s.size ), m]
      next if n.is_leaf?
      
      tmp = [] 
      tmp << n.offset
      n.each_child do |ch|
	tmp << ch.offset
      end
      tmp.sort!


      (o2l[ tmp[0] ] ... o2l[ tmp[tmp.size - 1] ]).each do |k|

	vm[k * 2][maxdepth - d - 1 ] = "%s%s" % [" " * (d2ml[d + 1] + 1), '|'] 
	vm[k * 2 + 1][maxdepth - d - 1] = "%s%s" % [" " * (d2ml[d + 1] + 1), '|'] 
      end

      vm[o2l[n.offset] * 2][maxdepth - d - 1] = "%s%s" % [" " * (d2ml[d + 1] + 1), '+'] 
      
    end

    vm.each do |vl|
      
      vl.each do |c|
	os.print c
      end
      os.puts
    end
    
  end

  def to_rule
  end

  def terminals
    tmp = []
    depth_first_traverse_pre_order do |n|
      tmp << n
    end
    return tmp.sort{|x,y|  x.sid <=> y.sid} 
  end

end
