require 'pos'

class HeadFinder < Hash
  class Rule 
    LEFT = 'left'
    RIGHT = 'right'
    
    attr_reader :label, :direction, :condition 
    def initialize(arg=nil) 
      @label = nil
      @direction = RIGHT
      @condition = []
      str2rule(arg) if arg != nil
    end
    
    def str2rule(str)
      return false if /^\#/ =~ str
      @label, @direction, condition = str.split(/[\s\t]+/,3)
      condition.split(/[\s\t]+/).each do |x|
	@condition << /^#{x}$/
      end 
      return self
    end

    def find (children)
      hid = 0
      
      @condition.each do |c|
	if @direction == RIGHT
	  i = children.size - 1 
	  while i >= 0
	    label = children[i].label
	    label = children[i].pos if children[i].leaf? 
	    label = label.best if label.kind_of?(PartOfSpeech)

	    if PartOfSpeech::PUNCTUATION =~ label
	      i -= 1
	      next 
	    end
	    
	    if c =~ '**'
	      return i 
	    end

	    if c =~ label
	      #puts ">>%s %s %s" % [i, c.source, label]
	      #puts "here %d" % [i]
	      return i 
	    end
	    i -= 1 
	  end
	  hid = children.size - 1 
	  label = children[hid].label
	  label = children[hid].pos if children[hid].leaf? 
	  label = label.best if label.kind_of?(PartOfSpeech)
	  while hid > 0 && PartOfSpeech::PUNCTUATION =~ label 
	    hid -= 1
	    label = children[hid].label
	    label = children[hid].pos if children[hid].leaf? 
	    label = label.best if label.kind_of?(PartOfSpeech)
	  end

	elsif @direction == LEFT
	  children.each_index do |i|
	    label = children[i].label
	    label = children[i].pos if children[i].leaf? 
	    label = label.best if label.kind_of?(PartOfSpeech)
	    next if PartOfSpeech::PUNCTUATION =~ label


	    if c =~ '**'
	      return i 
	    end
	    if c =~ label
	      
	      return  i
	    end
	  end
	  hid = 0

	  label = children[hid].label
	  label = children[hid].pos if children[hid].leaf? 
	  label = label.best if label.kind_of?(PartOfSpeech)

	  while hid < children.size - 1 && PartOfSpeech::PUNCTUATION =~ label
	    hid += 1

	    label = children[hid].label
	    label = children[hid].pos if children[hid].leaf? 
	    label = label.best if label.kind_of?(PartOfSpeech)
	  end
	end
      end
      return hid 
    end

    def to_s
      tmp = [] 
      @condition.each do |x|
	tmp << x.source.sub(/^\^/, '')
      end
      return "%s\t%s\t%s" % [@label, @direction, tmp.join("\t")]
    end
  end
  

  def initialize
    super
  end

  def read(is=DATA)
    while line = is.gets
      hr = Rule.new(line)
      self[hr.label] = hr 
    end 
    return self
  end

  def write(os=$stdout)
    self.each do |label, rule|
      os.puts rule.to_s
    end
  end

  def find(label, children)
    return children.size - 1 if self[label] == nil 
    return self[label].find(children)
  end

  def find(label, children)

    if self[label] == nil 
      hid = children.size - 1 

      return hid if hid < 0 


      label = children[hid].label 
      label = children[hid].pos if children[hid].leaf?

      while hid > -1 && PartOfSpeech::PUNCTUATION =~ label

	hid -= 1 
	label = children[hid].label 
	label = children[hid].pos if children[hid].leaf?
      end
      return hid 
    end
    return self[label].find(children)
  end


end

if __FILE__ == $0
  hf = HeadFinder.new
  hf.read()
  hf.write
end

__END__
NP      right   (POS|NN|NNP|NNPS|NNS) NX JJR CD JJ JJS RB QP NP 
ADJP    right   NNS QP NN \$ ADVP JJ VBN VBG ADJP JJR NP JJS DT FW RBR RBS SBAR RB
ADVP    left    RB RBR RBS FW ADVP TO CD JJR JJ IN NP JJS NN
CONJP   left    CC RB IN
FRAG    left    .
INTJ    right   .
LST     left    LS :
NAC     right   (NN|NNS|NNP|NNPS) NP NAC EX \$ CD QP PRP VBG JJ JJS JJR ADJP FW
PP      left    IN TO VBG VBN RP FW
PRN     right   .
PRT     left    RP
QP      right   \$ IN NNS NN JJ RB DT CD NCD QP JJR JJS
RRC     left    VP NP ADVP ADJP PP
S       right   TO IN VP S SBAR ADJP UCP NP
SBAR    right   WHNP WHPP WHADVP WHADJP IN DT S SQ SINV SBAR FRAG
SBARQ   right   SQ S SINV SBARQ FRAG
SINV    right   VBZ VBD VBP VB MD VP S SINV ADJP NP
SQ      right   VBZ VBD VBP VB MD VP SQ
UCP     left    .
VP      left    VBD VBN MD VBZ VB VBG VBP VP ADJP NN NNS NP
WHADJP  right   CC WRB JJ ADJP
WHADVP  left    CC WRB
WHNP    right   WDT WP WP\$ WHADJP WHPP WHNP
WHPP    left    IN TO FW
NX      right   (POS|NN|NNP|NNPS|NNS) NX JJR CD JJ JJS RB QP NP 
