require 'tree'
require 'pos'

class PhraseStructureTree < Tree

  class NotBinaryTreeForm < Exception ; end 
  attr_accessor :label, :head, :nonhead, :score, :alternative, :pos, :lex, :mark, :semantic, :chunk

  def initialize(data=nil, parent=nil, prev=nil, _next=nil, children=[])
    super(nil)
    @label = nil 
    @semantic = nil 
    @mark = nil
    @lex   = nil
    @pos   = nil 
    @chunk = nil 
    @head  = nil 
    @nonhead = nil
    @score = 0.0
    @alternative = []
    str2node(data) if data.kind_of?(String)
  end

  def nonhead_first_traverse(pre_order, in_order, post_order)
    pre_order.call(self) if pre_order != nil 
    
    @children.each_index do |i|
      next if i == @head 
      @children[i].nonhead_first_traverse(pre_order, in_order, post_order) 
      in_order.call(self) if in_order != nil 
    end

    @children[@head].nonhead_first_traverse(pre_order, in_order, post_order) if @head != nil 

    post_order.call(self) if post_order != nil 
  end

  def nonhead_first_traverse_pre_order(&proc)
    nonhead_first_traverse(proc, nil, nil)
  end

  def nonhead_first_traverse_in_order(&proc)
    nonhead_first_traverse(nil, proc, nil)
  end

  def nonhead_first_traverse_post_order(&proc)
    nonhead_first_traverse(nil, nil, proc)
  end


  def head_first_traverse(pre_order, in_order, post_order)
    pre_order.call(self) if pre_order != nil 
    
    #puts "%s [%s] %s" % [self.to_s, @head, @children[@head].to_s] if @head != nil 
    #puts self.bracket
    # p @head
    
    @children[@head].head_first_traverse(pre_order, in_order, post_order) if @head != nil && ! leaf?

    in_order.call(self) if in_order != nil 
    @children.each_index do |i|
      next if i == @head 
      @children[i].head_first_traverse(pre_order, in_order, post_order) 
    end
    post_order.call(self) if post_order != nil 
  end

  def head_first_traverse_pre_order(&proc)
    head_first_traverse(proc, nil, nil)
  end

  def head_first_traverse_in_order(&proc)
    head_first_traverse(nil, proc, nil)
  end

  def head_first_traverse_post_order(&proc)
    head_first_traverse(nil, nil, proc)
  end


  def bracket(left=@@left, right=@@right, delimiter=@@delimiter)
    return "%s%s%s%s%s" % [left, @pos, delimiter, @lex, right]  if @label == nil && @pos != nil  && @lex != nil 
    tmp = [] 
    tmp << "%s%s%s" % [@mark, @label.to_s, @semantic] if @pos == nil && @lex == nil 
    @children.each do |x|
      tmp << x.bracket(left, right, delimiter)
    end
    return "%s%s%s" % [left, tmp.join(delimiter), right]
  end


  def head_word
    return @lex if @head == nil || leaf?
    return @children[@head].head_word
  end

  def head_offset 
    return @offset if @head == nil || leaf?
    # puts "CHECK %s %s" % [@children.size, @head ]
    return @children[@head].head_offset
  end

  def head_pos
    if @head == nil || leaf?
      return @pos if @pos != nil 
      return [] 
    end
    return @children[@head].head_pos
  end

  def to_s
    return "%s %s" % [@pos, @lex] if leaf?
    return "%s%s%s" % [@mark, @label.to_s, @semantic]
  end

  def leaf?
    return super && @label == nil && @pos != nil  && @lex != nil 
  end

  def clone # shallow copy 
    tmp = self.type.new()
    tmp.label = self.label.clone 
    tmp.pos = self.pos.clone
    tmp.lex = self.lex.clone
  end

  def unary?
    return true if @children.size == 1
  end

  def head_is_left?
    raise NotBinaryTreeForm unless @children.size == 2
    return @head != nil && @head == 0
  end

  def head_is_right?
    raise NotBinaryTreeForm unless @children.size == 2
    return @head != nil && @head == 1
  end

  def str2node(str)
    if /^(\S+)[\s\t]*$/ =~ str # non-terminal

      label = $1.to_s
      #puts label
      if /^(\S+?)([\=\-\|]\S+)/ =~ label
	
	label = $1.to_s
	@semantic = $2.to_s
      end
      
      @label = label

      return self
    end
    
    if /^(\S+)[\s\t]+(\S+)$/ =~ str # terminal
      @pos = PartOfSpeech.str2pos($1.to_s)
      @lex = $2.to_s

      return self
    end
    
    raise "%s :invalid format" % [str]

  end
  
end

