class PartOfSpeech

  BOS = 'BOS'
  EOS = 'EOS'
  GAP = '-NONE-'

  UNKNOWN = 'UNK'
  # PUNCTUATION = /^(:|\,|\.)$/
  PUNCTUATION = /^(\.|\,|\`\`|\'\'|\-LRB\-|\-RRB\-|:|\#)$/
  FUNCTIONAL_WORD = /^(IN|DT|PDT|RP|CC|MD|TO|WDT)$/
  CLOSED_WORD = /^(IN|DT|PDT|RP|CC|MD|TO|WDT|PRP$?)$/
  CONJUNCTION = /^CC$/
  class InvalidFormat < Exception ; end

  attr_accessor :score, :sem, :str
  protected :str
  def initialize(_str=nil, _sem=nil) 
    @score = 0.0
    @str = _str
    @sem = _sem
  end

  def to_s
    return @str
  end

  def PartOfSpeech.str2pos(str)
    
    if /^(\S+)\-(\S+)/ =~ str
      return PartOfSpeech.new($1, $2)
    elsif /^(\S+)/ =~ str
      return PartOfSpeech.new($1)
    else
      raise Initialize, "Invalid Format as part-of-speech string"
    end
    
  end

  def functional_word?
    return FUNCTIONAL_WORD =~ @str 
  end
  
  def closed_word?
    return CLOSED_WORD =~ @str 
  end
  
  def PartOfSpeech.bos
    return PartOfSpeech.new(BOS.to_s)
  end

  def PartOfSpeech.eos
    return PartOfSpeech.new(EOS.to_s)
  end

  def PartOfSpeech.unknown
    return PartOfSpeech.new(UNKNOWN.to_s)
  end

  def PartOfSpeech.gap
    return PartOfSpeech.new(GAP.to_s)
  end

  def bos?
    return @str == BOS
  end

  def gap?
    @str == GAP
  end

  def eos?
    return @str == EOS
  end

  def unknown?
    return @str == UNKNOWN
  end

  def conjunction?
    return CONJUNCTION =~ @str
  end

  def punctuation?
    return is_punctuation?
  end

  def is_punctuation?
    return PUNCTUATION =~ @str  
  end

  def is_noun?
    /^N/ =~ @str
  end

  def is_proper_noun?
    @str == 'NNP' || @str == 'NNPS'
  end

  def is_verb?
    /^V/ =~ @str
  end

  def == (arg)
    return nil if arg == nil || ! arg.kind_of?(PartOfSpeech)
    return @str == arg.str
  end

  def =~ (arg)
    return nil if arg == nil || ! arg.kind_of?(PartOfSpeech)

    if /^NN/ =~ @str && /^NN/ =~ arg.str
      return true
    end

    if /^VB/ =~ @str && /^VB/ =~ arg.str
      return true
    end

    if /^JJ/ =~ @str && /^JJ/ =~ arg.str
      return true
    end

    if /^RB/ =~ @str && /^RB/ =~ arg.str
      return true
    end

    if /^PRP/ =~ @str && /^PRP/ =~ arg.str
      return true
    end

    if /^WP/ =~ @str && /^WP/ =~ arg.str
      return true
    end

    if is_punctuation? && arg.is_punctuation?
      return true
    end

    return @str == arg.str
  end

end

if __FILE__ == $0
  pos1 = PartOfSpeech.str2pos("NN")
  pos2 = PartOfSpeech.str2pos("NN-SUBJ")
  bos = PartOfSpeech.bos
  eos = PartOfSpeech.eos
  unk = PartOfSpeech.unknown

  p pos1
  p pos2
  p bos
  p eos
  p unk

end


