require 'phrase_structure_parser/phrase_structure_tree'
require 'pos'

class PhraseStructureTree < Tree

  INTERMEDIATE_MARK = '@'
  HEAD_MARK = '#'
  BOS = 'BOS'
  EOS = 'EOS'
  
  attr_accessor :first, :last

  def initialize(data=nil, parent=nil, prev=nil, _next=nil, children=[])
    super(nil)
    @label = nil 

    @semantic = nil 
    @mark = nil
    @lex   = nil
    @pos   = nil 
    @head  = nil 
    @nonhead = nil
    @score = 0.0
    @alternative = []
    @first = false
    @last = false
    str2node(data) if data.kind_of?(String)
  end

  def PhraseStructureTree.bos
    _bos = PhraseStructureTree.new
    _bos.pos = PartOfSpeech.new(BOS.to_s) 
    _bos.lex = ""
    return _bos
  end


  def PhraseStructureTree.eos
    _eos = PhraseStructureTree.new
    _eos.pos = PartOfSpeech.new(EOS.to_s) 
    _eos.lex = ""
    return _eos
  end

  def bos?
    return @pos.to_s == BOS
  end

  def eos?
    return @pos.to_s == EOS
  end

  def intermediate?
    return @mark == INTERMEDIATE_MARK
  end

  def to_cfg_rule
    return nil if leaf? 

    left = "<%s>" % [@label]
    left = "<%s%s>" % [@label, @mark] if intermediate? 
    tmp = []
    @children.each do |ch|
      if ch.leaf?
	x = ch.pos
	x = '-SHARP-' if ch.pos == '#'
	x = '-DOLLAR-' if ch.pos == '$'
	tmp << x
      else
	x = "<%s>" % [ch.label]
	x = "<%s%s>" %  [ch.label, ch.mark] if ch.intermediate?
	tmp << x
      end
    end
    
    return "%s --> %s" % [left, tmp.join(" ")]
  end

  def bracket(left=@@left, right=@@right, delimiter=@@delimiter, sem=false, h=false)
    return "%s%s%s%s%s" % [left, @pos, delimiter, @lex, right]  if leaf?
    
    tmp = [] 
    s = ''
    m = ''
    s = @semantic if sem
    m = @mark if @mark == INTERMEDIATE_MARK 
    # puts "%s %s" % [sem, h]

    if h
      m = @mark 
      m = "%s%s" % [m, @head]
    end

    # puts "%s%s" % [, sem] 

    tmp << "%s%s%s" % [@label.to_s, m, s] if ! leaf?

    @children.each do |x|
      tmp << x.bracket(left, right, delimiter, sem, h)
    end
    return "%s%s%s" % [left, tmp.join(delimiter), right]
  end


  def symbol
    return @pos if leaf?
    _mark = ''
    _mark = @mark if intermediate?
    return "%s%s" % [@label, _mark]
  end

  def to_s()
    return "%s %s" % [@pos, @lex] if leaf?
    return "%s%s" % [@label.to_s, @mark]
  end

  def str2node(str)
    if /^(\S+)[\s\t]*$/ =~ str # non-terminal
      @mark = HEAD_MARK.to_s
      label = $1.to_s
      if /^(\S+?)([\=\-\|]\S+)/ =~ label
	label = $1.to_s
	@semantic = $2.to_s
      end
      
      @label = label
      #puts @label
      if /^(.+)(#{INTERMEDIATE_MARK}|#{HEAD_MARK})(\d+)$/ =~  @label
	#puts $2
	@mark = $2 if $2 != ""
	#puts ">>>%s %s %s" % [$1, $2, $3]
	@head = $3.to_i
	@label = $1
      elsif /^(.+)(#{INTERMEDIATE_MARK})$/ =~  @label
	@mark = $2 
	# puts ">>>%s %s %s" % [$1, $2, $3]
	@label = $1
      end

      return self
    end
    
    if /^(\S+)[\s\t]+(\S+)$/ =~ str # terminal
      
      @pos = $1.to_s
      @lex = $2.to_s
      return self
    end
    
    raise "%s :invalid format" % [str]
  end

end
