class Tree 
  class NoExistingChild < Exception ; end

  @@delimitor = ' '
  @@left = '('
  @@right = ')'

  attr_accessor :data, :parent, :prev, :next, :children , :offset
    
  def initialize(data=nil, parent=nil, prev=nil, _next=nil, children=[])
    @offset   = nil 
    @data     = data
    @parent   = parent
    @prev     = prev
    @next     = _next
    @children = children
    @_size = nil 
  end

  def size
    @_size = 0 
    self.depth_first_traverse_pre_order do |n|
      @_size += 1 
    end
    return @_size
  end

  def str2node(str)
    @data = str 
  end

  def to_s
    return "%s" % [@data]
  end

  def bracket(left=@@left, right=@@right, delimitor=@@delimitor)
    return self.to_s if self.leaf?
    tmp = [] 
    tmp << left
    tmp << self.to_s
    @children.each do |x|
      tmp << x.bracket(left, right, delimitor)
    end
    tmp << right
    return tmp.join(delimitor) 
  end

  def unary?
    return @children.size == 1 
  end

  def binary?
    return @children.size == 2 
  end

  def nary?
    return @children.size > 2
  end

  def leaf?
    return @children.size == 0
  end

  def root?
    return @parent == nil 
  end

  def first?
    return @prev == nil 
  end


  def last?
    return @next == nil 
  end

  def terminal?
    return leaf?
  end

  def index(&condition)
    idx = 0 
    @children.each do |ch|
      return idx if condition.call(ch)
      idx += 1
    end
    return nil
  end

  def insert(node, pos=-1)

    if pos < 0
      @children << node 
      node.parent = self
      if (n = @children.size) > 1
	node.prev = @children[n-2] 
	@children[n-2].next = node
      end
      return node
    end

    if @children.size <= pos
      return insert(node)
    end
	  
    @children[pos, 0] = [node] 
    
    node.parent = self
    unless pos - 1 < 0 
      node.prev = @children[pos - 1] 
      @children[pos - 1].next = node
    end
	  
    node.next = @children[pos+1] 
    @children[pos+1].prev = node 
      
    return node
  end

  def delete(pos=-1)
    if @children.size <= pos
      raise Tree::NoExistingChild, "can not delete %d-th child" % [pos]
    end

    i = pos
    i = 0 if pos < 0 
    t = @children.delete_at(i)
    t.parent = nil 
    return t if @children.size == 0 

    if i == 0 
      t.next = nil 
      @children[i].prev = nil 
    elsif i == @children.size
      t.prev = nil
      @children[i-1].next = nil 
    else
      t.prev = nil 
      t.next = nil 
      @children[i].prev = @children[i-1]
      @children[i-1].next = @children[i]
    end
    t.parent = nil 
    return t
  end




  def left_branch(i=self.children.size - 1, &proc)
    
    return nil if @children.size < 3 || i < 1
    right = @children[i] 
    self.delete(i)

    while i > 0 && @children.size > 1
      left = @children[i-1]


      self.delete(i-1)
      inode = self.type.new

      inode.insert(left)
      inode.insert(right)
      inode = proc.call(inode)
      right = inode
      
      i -= 1
    end


    self.insert(right, i)

    return right
  end

  def right_branch(i=0, &proc)
    s = i 
    num_of_ch  = @children.size
    return nil if num_of_ch < 3 || i > num_of_ch - 2

    left = @children[i] 
    self.delete(i)

    while @children.size > 1 && i < @children.size  
      right = @children[i]
      self.delete(i)
      inode = self.type.new

      inode.insert(left)
      inode.insert(right)
      inode = proc.call(inode)
      left = inode
      
    end

    self.insert(left, s)

    return left
  end

  def depth
    depth = 0 ;
    current = self ;
    while (! current.root?)
      depth += 1
      current = current.parent
    end
    return depth 
  end

  def max_depth
    max = 0 
    self.depth_first_traverse_pre_order{ |x|
      if x.is_leaf? 
	m = x.depth
	if  m > max 
	  max = m 
	end
      end
    }
    return max
  end

  def breadth_first_traverse(&proc)
    proc.call(self) if root?

    @children.each do |node|
      proc.call(node)
    end
    
    @children.each do |node|
      node.breadth_first_traverse(max_depth, &proc)
    end
  end

  def depth_first_traverse(pre_order, in_order, post_order)
    pre_order.call(self) if pre_order != nil
    self.children.each do |node|
      node.depth_first_traverse(pre_order, in_order, post_order)
      in_order.call(self) if in_order != nil && node.first? && ! self.leaf?
    end
    in_order.call(self) if (in_order != nil && self.leaf?)

    post_order.call(self) if post_order != nil
  end

  def depth_first_traverse_pre_order (&proc)
    depth_first_traverse(proc, nil, nil)
  end

  def depth_first_traverse_in_order (&proc)
    depth_first_traverse(nil, proc, nil)
  end

  def depth_first_traverse_post_order (&proc)
    depth_first_traverse(nil, nil, proc)
  end
  
end

if __FILE__ == $0

    
end




